

#include <stdint.h>

//nRF52
//#include "app_timer.h"
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "app_util_platform.h"
#include "nrf_log.h"
#include "nrf_delay.h"
#include "nrf_drv_twi.h"
#include "nrf_twi.h"
#include "nrf_twim.h"
#include "boards.h"

#include "i2c_control.h"

// TWI (I2C) definitions and variables


static volatile bool m_xfer_done = false; // Indicates if operation on TWI has ended

#define TWI_INSTANCE_ID     0  // TWI instance ID
static const nrf_drv_twi_t m_twi = NRF_DRV_TWI_INSTANCE(TWI_INSTANCE_ID); // TWI instance
static uint8_t m_sample; // Buffer for samples read from I2C slave
uint8_t send_data[5];

extern void idle_state_handle(void);

#ifndef SCL_PIN
#define SCL_PIN ARDUINO_SCL_PIN
#define SDA_PIN ARDUINO_SDA_PIN
#endif //SCL_IN

///////////////////////////////////////////
//////// TWI (I2C) FUNCTIONS //////////////
///////////////////////////////////////////

/**
 * @brief TWI events handler.
 * If we remove this event handler, then TWI will become blocking. 
 * With the handler in place, we have to make sure to block after every rx call
 * Or else check the data only when twi_handler registers evt_done
 */
ret_code_t nrf_transfer_status;
void twi_handler(nrf_drv_twi_evt_t const * p_event, void * p_context)
{
    switch (p_event->type)
    {
        case NRF_DRV_TWI_EVT_DONE:
            m_xfer_done = true;
            nrf_transfer_status = NRF_SUCCESS;
            break;

        case NRF_DRV_TWI_EVT_ADDRESS_NACK:
            NRF_LOG_INFO("ADDRESS_NACK in twi_handler");
            m_xfer_done = true;
            nrf_transfer_status = NRF_ERROR_NOT_FOUND;
            break;
				
        case NRF_DRV_TWI_EVT_DATA_NACK:
            NRF_LOG_INFO("DATA_NACK in twi_handler");
            m_xfer_done = true;
            nrf_transfer_status = NRF_ERROR_INVALID_PARAM;
            break;

        default:
            NRF_LOG_INFO("DEFAULT in twi_handler");
            m_xfer_done = true;
            break;
    }
}

/**
 * @brief TWI/I2C initialization.
 */
void twi_init (void)
{
    ret_code_t err_code;

    const nrf_drv_twi_config_t twi_config = {
       .scl                = SCL_PIN,
       .sda                = SDA_PIN,
       //Limited by the temperature sensor to 100khz
       .frequency          = NRF_TWI_FREQ_100K, 
       .interrupt_priority = APP_IRQ_PRIORITY_HIGH,
       .clear_bus_init     = true,
       .hold_bus_uninit    = NRF_GPIO_PIN_NOPULL
    };

    err_code = nrf_drv_twi_init(&m_twi, &twi_config, twi_handler, NULL);
    APP_ERROR_CHECK(err_code);

    nrf_drv_twi_enable(&m_twi);
}

void twi_low_power(void)
{
    nrf_drv_twi_disable( &m_twi );
    nrf_drv_twi_uninit( &m_twi );
    
    // Nordic work around to prevent burning excess power due to chip errata
    // Use 0x400043FFC for TWI0.
    *(volatile uint32_t *)0x40003FFC = 0;
    *(volatile uint32_t *)0x40003FFC;
    *(volatile uint32_t *)0x40003FFC = 1;

    // Initialize the TWI with the event handler set to NULL. Makes the read
    // and write operations blocking. Must be done each time because  of the 
    // power off of the peripheral done above.
    twi_init();
}

void twi_disable(void)
{
    nrf_drv_twi_disable( &m_twi );
    nrf_drv_twi_uninit( &m_twi );
    
    // Nordic work around to prevent burning excess power due to chip errata
    // Use 0x400043FFC for TWI0.
    *(volatile uint32_t *)0x40003FFC = 0;
    *(volatile uint32_t *)0x40003FFC;
    *(volatile uint32_t *)0x40003FFC = 1;

}


//Read from the TWI address
//addr = peripherial address
//subaddr = sub address i.e. battery charge %
//dataread = array of bytes that are read
//len = number of bytes to be read
//WARNING: this funciton is not overflow safe, do not make len longer than the number of bytes in your array
ret_code_t twi_write_read(uint8_t addr, uint8_t subaddr[], uint8_t subaddr_len, uint8_t data_read[], uint8_t data_len)
{
    ret_code_t err_code;
    uint32_t i;

    // send register address

    if(subaddr_len != 0)
    {
        m_xfer_done = false; i = 0;
        err_code = nrf_drv_twi_tx(&m_twi, addr, subaddr, subaddr_len, true);  
        APP_ERROR_CHECK(err_code);

        while(m_xfer_done == false && i++ < 50) idle_state_handle();
        if(nrf_transfer_status != NRF_SUCCESS)
        {
    		twi_low_power();
            return nrf_transfer_status;
        }
    }
    
    // read it back
    m_xfer_done = false; i = 0;
    err_code = nrf_drv_twi_rx(&m_twi, addr, data_read, data_len);  
    
    while(m_xfer_done == false && i++ < 50) idle_state_handle();
    if(nrf_transfer_status != NRF_SUCCESS)
    {
        return nrf_transfer_status;
    }
    APP_ERROR_CHECK(err_code);

    
    twi_low_power();
    return err_code;
}

//Read from the TWI address
//addr = peripherial address
//dataread = array of bytes that are read
//len = number of bytes to be read
//WARNING: this funciton is not overflow safe, do not make len longer than the number of bytes in your array
ret_code_t twi_read_only(uint8_t addr, uint8_t data_read[], uint8_t data_len)
{
    ret_code_t err_code;
    uint32_t i;

    // read data
    m_xfer_done = false; i = 0;
    err_code = nrf_drv_twi_rx(&m_twi, addr, data_read, data_len);  
    
    while(m_xfer_done == false && i++ < 50) idle_state_handle();
    twi_low_power();

    if(nrf_transfer_status != NRF_SUCCESS)
    {
        return nrf_transfer_status;
    }
    APP_ERROR_CHECK(err_code);

    return err_code;
}

//Write to the TWI address
//addr = peripherial address
//subaddr = sub address i.e. battery charge %
//datawrite = array of bytes that are written
//len = number of bytes to be written
//warning: function is not buffer overflow safe
ret_code_t twi_write(uint8_t addr, uint8_t subaddr_and_data[], uint8_t len)
{
    ret_code_t err_code;
    uint32_t i;

    // send data
    m_xfer_done = false; i = 0;
    err_code = nrf_drv_twi_tx(&m_twi, addr, subaddr_and_data, len, true);  
    APP_ERROR_CHECK(err_code);
    
    while(m_xfer_done == false && i++ < 50) idle_state_handle();
    twi_low_power();

    if(nrf_transfer_status != NRF_SUCCESS)
    {
        return nrf_transfer_status;
    }

    return err_code;
}

